from __future__ import absolute_import as _absolute_import

#! /usr/bin/env python
'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''



def plot(scenario,img_filename=None,resource_height=1.0,show_task_labels=True,color_prec_groups=False,hide_tasks=[],border_width_perc=0.02,task_colors=dict(),fig_size=(15,5)) :
	"""
	Plot the given solved scenario using matplotlib

	Args:
		scenario:    scenario to plot
		kind:        the MIP-solver to use: CPLEX, GLPK
		msg:         0 means no feedback (default) during computation, 1 means feedback
		lp_filename: if set, then a .lp file will be written here

	Returns:
		0: good
	"""
	try :
		import matplotlib.patches as patches, matplotlib.pyplot as plt
	except :
		raise Exception('ERROR: matplotlib is not installed')
	import random

	S = scenario
	# trivial connected components implementation to avoid
	# having to import other packages just for that
	def get_connected_components(edges) :
		comps = dict()
		for v,u in edges :
			if v not in comps and u not in comps :
				comps[v] = v
				comps[u] = v
			elif v in comps and u not in comps :
				comps[u] = comps[v]
			elif v not in comps and u in comps :
				comps[v] = comps[u]
			elif v in comps and u in comps and comps[v] != comps[u] :
				old_comp = comps[u]
				for w in comps :
					if comps[w] == old_comp :
						comps[w] = comps[v]
		# replace component identifiers by integers startting with 0
		values = list(comps.values())
		comps = { T : values.index(comps[T]) for T in comps }
		return comps

	tasks = [ T for T in S.tasks() if T not in hide_tasks ] #TODO: upload since hide_tasks might contain string
	tasks = S.tasks()
	
	# get connected components dict for coloring
	# each task is mapping to an integer number which corresponds
	# to its connected component
	edges = [ (str(T),str(T)) for T in tasks ]
	if color_prec_groups :
		edges += [ (str(T),str(T_)) for P in S.precs if P.kind == '<' or P.kind == '<='\
	                   for T in P.tasks() for T_ in P.tasks() \
                           if T in tasks and T_ in tasks ]
	comps = get_connected_components(edges)

	# color map
	colors = ['#7EA7D8','#A1D372','#EB4845','#7BCDC8','#FFF79A'] #pastel colors
	#colors = ['red','green','blue','yellow','orange','black','purple'] #basic colors
	colors += [ [ random.random() for i in range(3) ] for x in range(len(S.tasks())) ] #random colors
	# replace colors with fixed task colors		
	for T in task_colors : colors[comps[str(T)]] = task_colors[T]
	color_map = { T : colors[comps[T]] for T in comps }

	solution = S.solution()
	hide_tasks_str = [ str(T) for T in hide_tasks ]
	solution = [ (T,R,x,y) for (T,R,x,y) in solution if T not in hide_tasks_str ]

	# resources list incl null resource
	resources = sorted(list(set([ R for (T,R,x,y) in solution ])))

	# plot solution
	#fig = plt.figure()
	fig, ax = plt.subplots(1, 1, figsize=fig_size)
	#ax = fig.add_subplot(111, aspect='equal')
	for (T,R,x,x_) in solution :
		y = resources.index(R)*resource_height
		ax.add_patch(
		    patches.Rectangle(
			(x+border_width_perc*resource_height, y+border_width_perc*resource_height),       # (x,y)
			x_-x-2*border_width_perc*resource_height,         # width
			resource_height*(1-2*border_width_perc),   # height
			color = color_map[T],
		    )
		)
		if show_task_labels :
			plt.text(x+0.1*resource_height,y+0.1*resource_height,str(T),fontsize=14,color='black')	

	# format graph
	plt.title(str(S))	
	plt.yticks([ resource_height*x + resource_height/2.0 for x in range(len(resources)) ],resources)
	plt.ylim(0,resource_height*len(resources))
	plt.xlim(0,max([ x_ for (I,R,x,x_) in solution ]))
	if img_filename :
		fig.figsize=(1,1)
		plt.savefig(img_filename,dpi=200,bbox_inches='tight')
	else :
		plt.show()

	return 0





#! /usr/bin/env python
'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''


"""
This module contains plotters that accept the pyschedule scenario format
as input. Each plotter should correspond to a class which offers the
method "plot" that takes a scenario as minimal input. If the scenario
has been solved beforehand, then a viualization should be returnd (in some way).
"""


from . import matplotlib




#! /usr/bin/env python
from __future__ import absolute_import as _absolute_import

'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

import time, copy
import pulp as pl


def _solve_mip(mip,kind='CBC',time_limit=None,msg=0) :

	start_time = time.time()
	# select solver for pl
	if kind == 'CPLEX' :
		mip.solve(pl.CPLEX_CMD(msg=msg,timelimit=time_limit))
	elif kind == 'GLPK' :
		mip.solve(pl.GLPK_CMD(msg=msg))
	elif kind == 'CBC' :
		options = []
		if time_limit :
			options += ['sec',str(time_limit)]
		mip.solve(pl.PULP_CBC_CMD(msg=msg,options=options))
	else :
		raise Exception('ERROR: solver '+kind+' not known')

	if msg :
		print('INFO: execution time for solving mip (sec) = '+str(time.time()-start_time))
	if mip.status == 1 and msg :		
		print('INFO: objective = '+str(pl.value(mip.objective)))


def solve(scenario,big_m=10000,kind='CBC',time_limit=None,msg=0,return_copy=False) :
	"""
	Shortcut to continuous mip
	"""
	if return_copy :
		scenario = copy.deepcopy(scenario)
	ContinuousMIP().solve(scenario,big_m=big_m,kind=kind,time_limit=time_limit,msg=msg)
	return scenario


def solve_discrete(scenario,max_time_step,kind='CBC',time_limit=None,msg=0,return_copy=False) :
	"""
	Shortcut to discrete mip
	"""
	if return_copy :
		scenario = copy.deepcopy(scenario)
	DiscreteMIP().solve(scenario,max_time_step,kind=kind,time_limit=time_limit,msg=msg)
	return scenario


class ContinuousMIP(object) :
	"""
	An interface to the pulp MIP solver package, supported are CPLEX, GLPK, CBC
	"""

	def __init__(self) :
		self.scenario = None
		self.big_m = None
		self.mip = None
		self.x = None  # mip variables shortcut


	def build_mip_from_scenario(self,msg=0) :

		S = self.scenario
		BIG_M = self.big_m #TODO: GLPK has problems with large number
		mip = pl.LpProblem(str(S), pl.LpMinimize)

		# check for objective
		if not S.objective :
			if msg : print('INFO: use makespan objective as default')
			S.use_makespan_objective()

		# task variables
		x = dict()
		for T in S.tasks() :
			x[T] = pl.LpVariable(T,0)

			# fix variables if start is given
			if T.start :
				mip += x[T] == T.start

			# add assignment variable for each possible resource
			for R in T.resources_req.resources() :
				x[(T,R)] = pl.LpVariable((T,R),0,1,cat=pl.LpBinary)

		# objective
		mip += sum([ x[T]*S.objective[T] for T in S.objective ])
		
		# everybody is required on one resource from each or clause
		for T in S.tasks() :
			for RA in T.resources_req :
				mip += sum([ x[(T,R)] for R in RA ]) >= 1

		# resource capacity constraints
		for R in S.resources() :
			if R.capacity :
				mip += sum([ x[(T,R)]*T.resources_req.capacity(R) for T in S.tasks() \
		                               if R in T.resources_req.resources() ]) <= R.capacity

		# same resource variable
		task_pairs = [ (T,T_) for T in S.tasks() for T_ in S.tasks() if str(T) < str(T_) ]
		for (T,T_) in task_pairs :
			shared_resources = list( set(T.resources_req.resources()) & set(T_.resources_req.resources()) )
			if shared_resources :
				x[(T,T_,'SameResource')] = pl.LpVariable((T,T_,'SameResource'),lowBound=0)#,cat=pl.LpInteger)
				x[(T_,T,'SameResource')] = pl.LpVariable((T_,T,'SameResource'),lowBound=0)#,cat=pl.LpInteger)
				mip += x[(T,T_,'SameResource')] == x[(T_,T,'SameResource')]
				for R in shared_resources :
					mip += x[(T,R)] + x[(T_,R)] - 1 <= x[(T,T_,'SameResource')]
				# ordering variables
				x[(T,T_)] = pl.LpVariable((T,T_),0,1,cat=pl.LpBinary)
				x[(T_,T)] = pl.LpVariable((T_,T),0,1,cat=pl.LpBinary)
				mip += x[(T,T_)] + x[(T_,T)] == 1

				mip += x[T] + T.length <= x[T_] + (1-x[(T,T_)])*BIG_M + (1-x[(T,T_,'SameResource')])*BIG_M 
				mip += x[T_] + T_.length <= x[T] + x[(T,T_)]*BIG_M + (1-x[(T,T_,'SameResource')])*BIG_M
				
		# precedence constraints
		for P in S.precs() :
			mip += x[P.left] + P.left.length + P.offset <= x[P.right] 
		
		# tight precedence constraints
		for P in S.precs_tight() :
			mip += x[P.left] + P.left.length + P.offset == x[P.right] 

		# conditional precedence constraints
		for P in S.precs_cond() :
			#T,T_ = tuple(sorted([P.left,P.right],key = lambda x : str(x) ))
			mip += x[P.left] + P.left.length + P.offset <= x[P.right] + \
	                        (1-x[(P.left,P.right)])*BIG_M + (1-x[(P.left,P.right,'SameResource')])*BIG_M

		# upper bounds
		for P in S.precs_up() :
			mip += x[P.left] + P.left.length <= P.right

		# lower bounds
		for P in S.precs_low() :
			mip += x[P.left] >= P.right

		self.mip = mip
		self.x = x


	def read_solution_from_mip(self,msg=0) :
		for T in self.scenario.tasks() :
			T.start = self.x[T].varValue
			T.resources = [ R for R in T.resources_req.resources() if self.x[(T,R)].varValue > 0 ]


	def solve(self,scenario,big_m=10000,kind='CBC',time_limit=None,msg=0) :
		"""
		Solves the given scenario using a continous MIP via the pulp package

		Args:
			scenario:    scenario to solve
			kind:        MIP-solver to use: CPLEX, GLPK, CBC
			big_m :      a large number to allow a big-m type model
			time_limit:  a time limit, only for CPLEX and CBC
			msg:         0 means no feedback (default) during computation, 1 means feedback
	
		Returns:
			scenario is solving was successful
			None if solving was not successful
		"""
		self.scenario = scenario
		self.big_m = big_m
		self.build_mip_from_scenario(msg=msg)
		_solve_mip(self.mip, kind=kind, time_limit=time_limit, msg=msg)

		if self.mip.status != 1 :
			if msg : print('ERROR: no solution found')
			#return None #TODO: problem sometimes still returned 0 when solution found

		self.read_solution_from_mip(msg=msg)
		return self.scenario


class DiscreteMIP(object) :
	"""
	pulp with time discretisation
	"""

	def __init__(self) :
		self.scenario = None
		self.max_time_step = None
		self.mip = None
		self.x = None  # mip variables shortcut


	def build_mip_from_scenario(self,msg=0) :
		S = self.scenario
		MAX_TIME = self.max_time_step
		mip = pl.LpProblem(str(S), pl.LpMinimize)

		# check for objective
		if not S.objective :
			if msg : print('INFO: use makespan objective as default')
			S.use_makespan_objective()
		
		# task variables
		x = dict()
		cons = list()
		for T in S.tasks() :
			for t in range(MAX_TIME) :
				x[(T,t)] = pl.LpVariable((T,t),0,1)
			# lower boundary conditions
			cons.append( pl.LpConstraint( x[(T,0)], sense=0, rhs=1 ) )
			cons.append( pl.LpConstraint( x[(T,self.max_time_step-1)], sense=0, rhs=0 ) ) #required for no solution feedback
	
			# fix variables if start is given
			if T.start :
				if T.start > self.max_time_step :
					raise Exception('ERROR: fixed start of task '+str(T)+' larger than max time step '+str(max_time_step))
				cons.append( pl.LpConstraint( x[(T,T.start)], sense=0, rhs=1 ) )

			# generate variabels for task resources
			for R in T.resources_req.resources() :
				# variables
				for t in range(MAX_TIME) :
					x[(T,R,t)] = pl.LpVariable((T,R,t),0,1,cat=pl.LpBinary)

				# monotonicity
				for t in range(MAX_TIME-1) :
					cons.append( pl.LpConstraint( pl.LpAffineExpression([ (x[(T,R,t)],1), (x[(T,R,t+1)],-1) ]), sense=1, rhs=0 ) )

			# consider each resource selection
			for RA in T.resources_req :
				# sum up resource distribution for each resource selection
				for t in range(MAX_TIME) :
					affine = pl.LpAffineExpression([ (x[(T,R,t)],1) for R in RA ] + [(x[T,t],-1)])
					cons.append( pl.LpConstraint( affine, sense=0, rhs=0 ) )

		# objective
		mip += pl.LpAffineExpression([ ( x[(T,t)], S.objective[T]) for T in S.objective for t in range(MAX_TIME)  ] )

		# resource non-overlapping constraints 
		for R in S.resources() :
			for t in range(MAX_TIME-1) :
				resource_tasks = [ T for T in S.tasks() if R in T.resources_req.resources() ]
				affine = pl.LpAffineExpression([ (x[(T,R,max(t-T.length,0))], 1) for T in resource_tasks ] + \
                                                                 [ (x[(T,R,t)], -1) for T in resource_tasks ])
				con =  pl.LpConstraint( affine, sense=-1, rhs=1.0 )
				cons.append(con)
				#prob += sum([ x[(T,max(t-T.length,0))]-x[(T,t)] for T in resource_tasks ]) <= 1.0

			if R.capacity :
				resource_tasks = [ T for T in S.tasks() if R in T.resources_req.resources() ]
				affine = pl.LpAffineExpression([ (x[T,R,0],T.resources_req.capacity(R)) for T in resource_tasks ])
				con =  pl.LpConstraint( affine, sense=-1, rhs=R.capacity )
				cons.append(con)

		# precedence constraints
		for P in S.precs() :
			for t in range(MAX_TIME) :
				affine = pl.LpAffineExpression([ ( x[(P.left,t)], 1), (x[(P.right,min(t+P.left.length,MAX_TIME-1))],-1)  ])
				con = pl.LpConstraint( affine, sense=-1, rhs=0 )
				cons.append(con)

		# tight precedence constraints
		for P in S.precs_tight() :
			for t in range(MAX_TIME) :
				affine = pl.LpAffineExpression([ ( x[(P.left,t)], 1), (x[(P.right,min(t+P.left.length,MAX_TIME-1))],-1)  ])
				con = pl.LpConstraint( affine, sense=0, rhs=0 )
				cons.append(con)

		# conditional precedence constraints
		for P in S.precs_cond() :
			shared_resources = list( set(P.left.resources_req.resources()) & set(P.right.resources_req.resources()) )
			for R in shared_resources :	
				for t in range(MAX_TIME) :
					affine = pl.LpAffineExpression([ ( x[(P.left,R,max(t-P.left.length,0) )],1 ), (x[(P.left,R,t)],-1),\
		                                                          ( x[(P.right,R,max(t-P.left.length,0))], 1), (x[(P.right,R,min(t+P.offset,MAX_TIME-1) )],-1) ])
					con = pl.LpConstraint( affine, sense=-1, rhs=1 )
					cons.append(con)

		# lower bounds
		for P in S.precs_low() :
			cons.append( x[(P.left,P.right)] == 1 )

		# upper bounds
		for P in S.precs_up() :
			cons.append( x[(P.left,P.right)] == 0 )

		for con in cons :
			mip.addConstraint(con)

		self.mip = mip
		self.x = x


	def read_solution_from_mip(self,msg=0) :
		for T in self.scenario.tasks() :
			T.start = max([ t for t in range(self.max_time_step) if self.x[(T,t)].varValue == 1.0 ])
			T.resources = [ R for R in T.resources_req.resources() if self.x[(T,R,0)].varValue > 0 ]


	def solve(self,scenario,max_time_step='100',kind='CBC',time_limit=None,msg=0) :
		"""
		Solves the given scenario using a discrete MIP via the pulp package

		Args:
			scenario:            scenario to solve
			kind:                MIP-solver to use: CPLEX, GLPK, CBC
			max_time_step :     the number of time steps to model
			time_limit:         a time limit, only for CPLEX and CBC
			msg:                 0 means no feedback (default) during computation, 1 means feedback
	
		Returns:
			scenario is solving was successful
			None if solving was not successful
		"""
		self.scenario = scenario
		self.max_time_step = max_time_step
		self.build_mip_from_scenario(msg=msg)
		_solve_mip(self.mip,kind=kind,time_limit=time_limit,msg=msg)

		if self.mip.status != 1 :
			if msg : print ('ERROR: no solution found')
			return None
		self.read_solution_from_mip(msg=msg)	
		return self.scenario ##TODO: check what to return
	












#! /usr/bin/env python
'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

import time, os

#TODO: add capacity constraints (with numbers)
#TODO: add alternative resources

def solve(scenario,msg=0) :

	S = scenario
	solvers_path = os.path.dirname(os.path.realpath(__file__))
		
	# no objective specified
	if not S.objective :
		S.use_makespan_objective()

	# check if there are alternative resources
	if max([ len(RA) for T in S.tasks() for RA in T.resources_req ]) > 1 :
		raise Exception('ERROR: alternative resources not supported yet for CP-optimizer')

	resources = S.resources()
	tasks = S.tasks()
	precedences = S.precs_lax()

	id_to_resource = dict(zip(range(len(resources)),resources))
	resource_to_id = dict(zip(resources,range(len(resources))))

	id_to_task = dict(zip(range(len(tasks)),tasks))
	task_to_id = dict(zip(tasks,range(len(tasks))))

	Resources = [ (resource_to_id[R],R.capacity) for R in resources if R.capacity != None ]
	Resources += [ (resource_to_id[R],1000) for R in resources if R.capacity == None ] #TODO: remove 100
	Resources = sorted(Resources)

	# Tasks for .dat-file
	Tasks = list()
	TaskResources = list()
	for T in tasks :				
		Tasks.append( (task_to_id[T],int(T.length)) )
		for RA in T.resources_req :
			if len(RA) != 1 :
				raise Exception('ERROR: opl can only deal with single resource reqs')
			for R in RA :
				TaskResources.append( (task_to_id[T],resource_to_id[R]) )
					
	Tasks = sorted(Tasks)
	TaskResources = sorted(TaskResources)

	# Precedences for .dat-file
	Precedences = list()
	for P in S.precs_lax() :
		Precedences.append(( task_to_id[P.left],task_to_id[P.right],P.offset))

	TightPrecedences = list()
	for P in S.precs_tight() :
		Precedences.append(( task_to_id[P.left],task_to_id[P.right],P.offset))

	CondPrecedences = list()
	for P in S.precs_cond() :
		CondPrecedences.append(( task_to_id[P.left],task_to_id[P.right],P.offset))

	LowerBounds = list()
	for P in S.precs_up() :
		UpperBounds.append((task_to_id[P.left],offset))

	UpperBounds = list()
	for P in S.precs_low() :
		LowerBounds.append((task_to_id[P.right],-offset))

	# TODO: support all other precedences. Support also tight upper and lower bounds

	Precedences = sorted(Precedences)
	CondPrecedences = sorted(CondPrecedences)

	# add objective
	Objectives = list()
	for key in S.objective :
		Objectives.append( (task_to_id[key],S.objective[key]) )
	Objectives = sorted(Objectives)

	# function to convert table into opl type typle lists
	def to_str(l) :
		return '\n'.join([ '<'+' '.join([ str(x) for x in row ])+'>'  \
                                  for row in l ])

	if not os.path.exists(solvers_path+'/tmp') :
		os.makedirs(solvers_path+'/tmp')

	# write .dat-file
	f = open(solvers_path+'/tmp/cpoptimizer.dat','w')
	f.write('Objectives={\n'+to_str(Objectives)+'\n};\n\n')
	f.write('Tasks={\n'+to_str(Tasks)+'\n};\n\n')
	f.write('Resources={\n'+to_str(Resources)+'\n};\n\n')
	f.write('TaskResources={\n'+to_str(TaskResources)+'\n};\n\n')
	f.write('Precedences={\n'+to_str(Precedences)+'\n};\n\n')
	f.write('TightPrecedences={\n'+to_str(TightPrecedences)+'\n};\n\n')
	f.write('CondPrecedences={\n'+to_str(CondPrecedences)+'\n};\n\n')
	f.write('UpperBounds={\n'+to_str(UpperBounds)+'\n};\n\n')
	f.write('LowerBounds={\n'+to_str(LowerBounds)+'\n};\n\n')
	f.close()

	# run cp-optimizer
	start_time = time.time()
	os.system('oplrun '+solvers_path+'/cpoptimizer.mod '+solvers_path+'/tmp/cpoptimizer.dat')
	if msg : print('INFO: execution time (sec) = '+str(time.time()-start_time))

	# parse output 
	from pyparsing import Literal,Word,alphas,nums,printables,OneOrMore,ZeroOrMore,dblQuotedString,Group
	
	INT = Word( nums )

	int_row = Group( Literal("<").suppress() + \
	                    INT + INT + INT + INT + \
                            Literal(">").suppress() )

	res_seq_row = Group( Literal("<").suppress()  + dblQuotedString + \
	                    INT + INT + INT + \
	                    INT + INT + INT + \
                            Literal(">").suppress() )
	res_seq = Group( Literal("{").suppress() + ZeroOrMore(res_seq_row) + Literal("}").suppress() )

	plan = Group(ZeroOrMore(int_row)) + Group(ZeroOrMore(res_seq))

	opl_plan = plan.parseFile(solvers_path+'/tmp/cpoptimizer.out')

	int_plan = opl_plan[0]
	res_seq_plan = opl_plan[1]

	# get starts of tasks
	starts = { i : int(int_plan[i][1]) for i in range(len(int_plan)) }

	assign = { int(row[2]) : id_to_resource[i] for i in range(len(res_seq_plan)) \
                                  for row in res_seq_plan[i] if int(row[4]) == starts[int(row[2])] }

	# add to scenario
	for T in S.tasks() :
		T.start = starts[task_to_id[T]] #second column is start
		T.resources = [assign[task_to_id[T]]] #assume only one resource









#! /usr/bin/env python
from __future__ import print_function

'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

"""
python package to formulate and solve resource-constrained scheduling problems
"""

#TODO: cpoptimier with cond precedence constraints


from collections import OrderedDict as _DICT_TYPE

try: # allow Python 2 vs 3 compatibility
	_maketrans = str.maketrans
except AttributeError:
	from string import maketrans as _maketrans

def OR(L) :
	"""
	method to iterate the or-operator over a list to allow lists of alternative resources, e.g. OR([R1,R2,R3]) = R1 | R2 | R3
	"""
	return reduce(lambda x, y : x | y, L)

def _isnumeric(var) :
	return isinstance(var,(int)) # only integers are accepted

def _isiterable(var) :
	return isinstance(var,(list,set,tuple)) #restriction to list,set,tuple to avoid funny cases



class _SchedElement(object) : # extend string object

	def __init__(self,name,numeric_name_prefix='E') :
		# purely numeric names are not allowed
		if _isnumeric(name) :
			name = numeric_name_prefix+str(name)
		# remove illegal characters
		trans = _maketrans("-+[] ->/","________")
		self.name = str(name).translate(trans)

	def __repr__(self) :
		return self.name

	def __str__(self) :
		return self.name

	def __hash__(self) :
		return self.name.__hash__()



class _SchedElementAffine(_DICT_TYPE) :

	def __init__(self,unknown=None,element_class=None,kind='+') :
		_DICT_TYPE.__init__(self)
		self.element_class = element_class
		self.kind = kind

		if unknown == None :
			pass
		elif isinstance(unknown,self.element_class) :
			self[unknown] = 1
		elif _isnumeric(unknown) :
			self[1] = unknown
		elif isinstance(unknown,type(self)) :
			self.update(unknown)
		elif isinstance(unknown,list) :
			self.update(_DICT_TYPE(unknown))
		else :
			raise Exception('ERROR: cant init '+str(self)+' from '+str(unknown))

	def __add__(self,other) :
		if _isiterable(other) :
			return [ self + x for x in other ]
		if isinstance(other,type(self)) :
			new = type(self)(self)
			for key in other :
				if key in new :
					new[key] += other[key]
				else :
					new[key] = other[key]
			return new
		else :
			new = type(self)(other)
			return self + new

	def __sub__(self,other) :
		if _isiterable(other) :
			return [ self - x for x in other ]
		if isinstance(other,type(self)) :
			new = type(self)(self)
			for key in other :
				if key in new :
					new[key] -= other[key]
				else :
					new[key] = -other[key]
			return new
		else :
			return self - type(self)(other)

	def __mul__(self,other) :
		if isinstance(other,type(self)) :
			new = type(self)(self)
			if 1 in other :
				for key in self :
					new[key] *= other[1]
			else :
				raise Exception('ERROR: multipy '+str(self)+' only with number \
                                                 but not with '+str(other))
			return new
		else :
			return self * type(self)(other)

	def __repr__(self) :
		def format_coeff(val) :
			if val != 1 :
				return str(val)+'x'
			else :
				return ''
		return (' '+self.kind+' ').join([ format_coeff(self[key])+str(key) for key in self ])

	def __hash__(self) :
		return self.__repr__().__hash__()


	
class Scenario(_SchedElement):
	
	def __init__(self,name='scenario not named') :
		_SchedElement.__init__(self,name,numeric_name_prefix='S')
		self.objective = _TaskAffine()
		self.T = _DICT_TYPE() #tasks
		self.R = _DICT_TYPE() #resources
		self.constraints = list()
		#self.precs = list()

	def Task(self,name,length=1) :
		"""
		Adds a task with the given name to the scenario. Task names need to be unique.
		"""
		T = Task(name,length)
		if str(T) not in self.T : #[ str(T_) for T_ in self.tasks ] :
			self.T[str(T)] = T
		else :
			raise Exception('ERROR: task with name '+str(T)+' already contained in scenario '+str(self))
		return T

	def tasks(self) :
		return self.T.values()

	def Resource(self,name,capacity=None) :
		"""
		Adds a resource with the given name to the scenario. Resource names need to be unique.
		"""
		R = Resource(name,capacity)
		if str(R) not in self.R : #[ str(R_) for R_ in self.resources ] :
			self.R[str(R)] = R #.append(R)
		else :
			raise Exception('ERROR: resource with name '+str(R)+' already contained in scenario '+str(self))
		return R

	def resources(self) :
		return self.R.values()

	def R(self,name) :
		return self._resources[name]

	def solution(self) :
		"""
		Returns the last computed solution in tabular form with columns: task, resource, start, end
		"""
		solution = [ (str(T),str(R),T.start,T.start+T.length) \
                             for T in self.tasks() for R in T.resources ]
		solution += [ (str(T),'',T.start,T.start+T.length) \
                             for T in self.tasks() if not T.resources_req.resources() ]
		solution = sorted(solution, key = lambda x : (x[1],x[2])) # sort according to resource and start
		return solution

	def use_makespan_objective(self) :
		"""
		Set the objective to the makespan of all included tasks without a fixed start
		"""
		self.objective.clear()
		tasks = list(self.tasks())
		makespan = self.Task('MakeSpan')
		makespan += self.Resource('MakeSpan')
		self += makespan*1
		for T in tasks :
			if not T.start :
				self += T < makespan

	def clear_task_starts(self) :
		"""
		Remove the start times of all tasks
		"""
		for T in self.tasks() :
			T.start = None

	def precs(self) :
		return [ C for C in self.constraints if isinstance(C,Precedence) ]

	def precs_tight(self) :
		return [ C for C in self.constraints if isinstance(C,PrecedenceTight) ]

	def precs_cond(self) :
		return [ C for C in self.constraints if isinstance(C,PrecedenceCond) ]

	def precs_low(self) :
		return [ C for C in self.constraints if isinstance(C,PrecedenceLow) ]

	def precs_up(self) :
		return [ C for C in self.constraints if isinstance(C,PrecedenceUp) ]

	def __iadd__(self,other) :
		if _isiterable(other) :
			for x in other : self += x
			return self

		elif isinstance(other,_TaskAffineConstraint) :
			pos_tasks = [ T for T in other if isinstance(T,Task) and other[T] >= 0 ]
			neg_tasks = [ T for T in other if isinstance(T,Task) and other[T] < 0 ]
			if len(neg_tasks) > 1 or len(pos_tasks) > 1 :
				raise Exception('ERROR: can only deal with simple precedences of \
		                                 the form T1 + 3 < T2 or T1 < 3 and not '+str(other) )

			# get offset
			offset = 0
			if 1 in other : offset = other[1]

			if pos_tasks and neg_tasks :
				left = pos_tasks[0]
				right = neg_tasks[0]
				if other.kind == '<' :
					self.constraints.append(Precedence(left=left,offset=offset,right=right))
				elif other.kind == '<=' :
					self.constraints.append(PrecedenceTight(left=left,offset=offset,right=right))
				elif other.kind == '<<' :
					shared_resources = list( set(left.resources_req.resources()) & set(right.resources_req.resources()) )
					prec = PrecedenceCond(left=left,offset=offset,right=right)
					if not shared_resources :
						raise Exception('ERROR: tried to add precedence '+str(prec)+' but tasks dont compete for resources')
					self.constraints.append(prec)
				return self
			elif pos_tasks and not neg_tasks :
				left = pos_tasks[0]
				right = -offset
				self.constraints.append(PrecedenceUp(left=left,right=right))
				return self
			elif not pos_tasks and neg_tasks :
				left = neg_tasks[0]
				right = offset
				self.constraints.append(PrecedenceLow(left=left,right=right))
				return self

			raise Exception('ERROR: cannot add constraint '+str(other)+' to scenario')

		elif isinstance(other,(Task,_TaskAffine)) : #TODO: attention Precedence is also _TaskAffine
			self.objective += other
			return self

		raise Exception('ERROR: cant add '+str(other)+' to scenario '+str(self))

	def __repr__(self) :
		s = '\n##############################################\n\n'
		s += 'SCENARIO: '+str(self)+'\n\n'
		#print objective
		s += 'OBJECTIVE: '+str(self.objective)+'\n\n'

		s += 'RESOURCES:\n'
		for R in self.resources() :
			s += str(R)+'\n'
		s += '\n'
		# print tasks
		s += 'TASKS:\n'
		s += '\n'.join([ str(T)+' : '+ str(T.resources_req) for T in sorted(self.tasks()) ]) + '\n\n'
		# print resources

		if self.precs() :
			# print precedences
			s += 'PRECEDENCES:\n'
			s += '\n'.join([ P.__repr__() for P in self.precs() ]) + '\n'
			s += '\n'

		if self.precs_tight() :
			# print precedences
			s += 'TIGHT PRECEDENCES:\n'
			s += '\n'.join([ P.__repr__() for P in self.precs_tight() ]) + '\n'
			s += '\n'

		if self.precs_cond() :
			# print precedences
			s += 'COND PRECEDENCES:\n'
			s += '\n'.join([ P.__repr__() for P in self.precs_cond() ]) + '\n'
			s += '\n'

		if self.precs_low() :
			# print precedences
			s += 'LOWER BOUNDS:\n'
			s += '\n'.join([ P.__repr__() for P in self.precs_low() ]) + '\n'
			s += '\n'

		if self.precs_up() :
			# print precedences
			s += 'UPPER BOUNDS:\n'
			s += '\n'.join([ P.__repr__() for P in self.precs_up() ]) + '\n'
			s += '\n'

		s += '##############################################\n'
		return s



class Task(_SchedElement) :
	"""
	A task to be processed by at least one resource
	"""

	def __init__(self,name,length=1,start=None,resources=None) :
		_SchedElement.__init__(self,name,numeric_name_prefix='T')
		self.start = start
		self.length = length
		self.resources = resources
		self.resources_req = _ResourceReq()

	def __len__(self) :
		return self.length

	def __iadd__(self,other) :
		if _isiterable(other) :
			for x in other : self += x		
		elif isinstance(other,(Resource,_ResourceAffine,_ResourceReq)) :
			self.resources_req += other
		else :
			raise Exception('ERROR: cant add '+str(other)+' to task '+str(self))
		return self

	def __lt__(self,other) :
		return _TaskAffine(self) < other

	def __gt__(self,other) :
		return _TaskAffine(self) > other

	def __le__(self,other) :
		return _TaskAffine(self) <= other

	def __ge__(self,other) :
		return _TaskAffine(self) <= other

	def __ne__(self,other) :
		return _TaskAffine(self) != other

	def __lshift__(self,other) :
		return _TaskAffine(self) << other

	def __rshift__(self,other) :
		return _TaskAffine(self) >> other

	def __add__(self,other) :
		return _TaskAffine(self) + other

	def __sub__(self,other) :
		return _TaskAffine(self) - other

	def __mul__(self,other) :
		return _TaskAffine(self) * other



class _TaskAffine(_SchedElementAffine) :

	def __init__(self,unknown=None) :
		_SchedElementAffine.__init__(self,unknown=unknown,element_class=Task)

	def __lt__(self,other) :
		if _isiterable(other) :
			return [ self < x for x in other ]
		if not isinstance(other,type(self)) :
			return self < _TaskAffine(other)
		return _TaskAffineConstraint(self-other,'<')

	def __gt__(self,other) :
		if _isiterable(other) :
			return [ self > x for x in other ]
		if not isinstance(other,type(self)) :
			return self > _TaskAffine(other)
		return _TaskAffine(other) < self

	def __le__(self,other) :
		if _isiterable(other) :
			return [ self <= x for x in other ]
		if not isinstance(other,type(self)) :
			return self <= _TaskAffine(other)
		return _TaskAffineConstraint(self-other,'<=')

	def __ge__(self,other) :
		if _isiterable(other) :
			return [ self >= x for x in other ]
		if not isinstance(other,type(self)) :
			return self >= _TaskAffine(other)
		return _TaskAffine(other) <= self

	def __lshift__(self,other) :
		if _isiterable(other) :
			return [ self << x for x in other ]
		if not isinstance(other,type(self)) :
			return self << _TaskAffine(other)
		return _TaskAffineConstraint(self-other,'<<')

	def __rshift__(self,other) :
		if _isiterable(other) :
			return [ self >> x for x in other ]
		if not isinstance(other,type(self)) :
			return self >> _TaskAffine(other)
		return _TaskAffine(other) << self



class _TaskAffineConstraint(_TaskAffine) :
	"""
	A representation of some inequality of e.g. the form T1 + T2*3 + 2 < 0
	The inquality sign is determined by parameter kind. 
	"""
	def __init__(self,task_affine,kind) :
		_TaskAffine.__init__(self)
		self.update(_DICT_TYPE(task_affine))
		self.kind = kind



class _Constraint(_SchedElement) :
	"""
	An arbitrary constraint
	"""
	def __init__(self) :
		_SchedElement.__init__(self,name='',numeric_name_prefix='C')



class Precedence(_Constraint) :
	"""
	A precedence constraint of two tasks, left and right, and an offset, e.g. T1 + 3 < T2
	"""
	def __init__(self,left,offset,right,kind='<') :
		_Constraint.__init__(self)
		self.left = left
		self.right = right
		self.offset = offset
		self.kind = kind

	def tasks(self) :
		return [self.left,self.right]

	def __repr__(self) :
		s = str(self.left) + ' '
		if self.offset != 0 :
			s += '+ ' + str(self.offset) + ' '
		s += str(self.kind) + ' ' + str(self.right)
		return s

	def __hash__(self) :
		return self.__repr__().__hash__()
		


class PrecedenceTight(Precedence) :
	"""
	A tight precedence of the form T1 + 3 <= T2, where 3 is the exact distance
	"""
	def __init__(self,left,offset,right) :
		Precedence(left,offset,right,kind='<=')



class PrecedenceCond(Precedence) :
	"""
	A conditional precedence of the form T1 + 3 << T2, where 3 is the changeover cost
	"""
	def __init__(self,left,offset,right) :
		Precedence.__init__(left,offset,right,kind='<<')



class PrecedenceUp(Precedence) :
	"""
	An upper bound of the form T1 < 5
	"""
	def __init__(self,left,right) :
		Precedence.__init__(left,0,right,kind='<')

	def tasks(self) :
		return [self.left]



class PrecedenceLow(Precedence) :
	"""
	An upper bound of the form T1 > 5
	"""
	def __init__(self,left,right) :
		Precedence.__init__(left,0,right,kind='<')

	def tasks(self) :
		return [self.left]
		


class Resource(_SchedElement) :
	"""
	A resource which can process at most one task per time step
	"""

	def __init__(self,name=None,capacity=None) :
		_SchedElement.__init__(self,name,numeric_name_prefix='R')
		self.capacity = capacity

	def __iadd__(self,other) :
		if isinstance(other,Task) :
			other += self
			return self
		elif isinstance(other,Precedence) :
			# check if each task in prec requires resource
			wrong_tasks = [ T for T in other.tasks() if not self in T.resources_req.resources()]
			if not wrong_tasks :
				self.precs.add(other)
			else :
				raise Exception('ERROR: Precedence '+str(other)+' includes task(s) '+\
                                                 ','.join([str(T) for T in wrong_tasks])+\
                                                 ' that do(es) not require resource '+str(self))
		elif _isiterable(other) :
			for x in other : self += x	
		else :
			raise Exception('ERROR: cant add '+str(other)+' to resource '+str(self))
		return self
		

	def __add__(self,other) :
		return _ResourceReq(_ResourceAffine(self)) + other

	def __mul__(self,other) :
		return _ResourceAffine(self) * other

	def __or__(self,other) :
		return _ResourceAffine(self) | other



class _ResourceAffine(_SchedElementAffine) :

	def __init__(self,unknown=None) :
		_SchedElementAffine.__init__(self,unknown=unknown,element_class=Resource,kind='|')

	def __or__(self,other) :
		return super(_ResourceAffine,self).__add__(_ResourceAffine(other)) #add of superclass

	def __add__(self,other) :
		return _ResourceReq(self) + other



class _ResourceReq(list) :

	def __init__(self,unknown=None) :
		list.__init__(self)
		if unknown :
			self.append(_ResourceAffine(unknown))

	def __add__(self,other) :
		if _isiterable(other) :
			for x in other : self += x
		other = _ResourceAffine(other)
		if set(other) & set(self.resources()) :
			raise Exception('ERROR: '+str(other)+' already somewhat contained in resource requirement '+str(self))
		self.append(other)
		return self

	def __iadd__(self,other) :
		self = self + other
		return self

	def __repr__(self) :
		return ' + '.join([ str(x) for x in self])

	def resources(self) :
		return [ R for RA in self for R in RA ]

	def capacity(self,R) :
		RA = [ RA_ for RA_ in self if R in RA_ ][0]
		return RA[R]
			




















from setuptools import setup, find_packages

setup(name='pyschedule',
      version='0.1.3.2',
      description='A python package to formulate and solve resource-constrained scheduling problems: flow- and job-shop, travelling salesman, vehicle routing and all kind of combinations',
      url='https://github.com/timnon/pyschedule',
      author='Tim Nonner',
      author_email='tim@nonner.de',
      license='Apache 2.0',
      packages=['pyschedule','pyschedule.solvers','pyschedule.plotters'],
      package_data={'': ['pyschedule.mod']},
      package_dir={'':'src'},
      include_package_data=True,
      install_requires=['pulp'])


